/*
    unicap
    Copyright (C) 2004  Arne Caspari

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __UNICAP_H__
#define __UNICAP_H__


#include <sys/types.h>
#include <sys/time.h>

#include <unistd.h>

#include <unicap_status.h>



enum _unicap_buffer_type
{
	UNICAP_BUFFER_TYPE_USER = 0, 
	UNICAP_BUFFER_TYPE_SYSTEM,
};

typedef enum _unicap_buffer_type unicap_buffer_type_t;


#define UNICAP_FLAGS_BUFFER_TYPE_USER   (UNICAP_BUFFER_TYPE_USER + 1)      
#define UNICAP_FLAGS_BUFFER_TYPE_SYSTEM (UNICAP_BUFFER_TYPE_SYSTEM << 1)

#define UNICAP_FLAGS_DEVICE_TYPE_REMOTE 2

struct _unicap_device
{
	char identifier[128]; // unique textual id of this device
	char model_name[128];
	char vendor_name[128];
		
	unsigned long long model_id; // serial number of device, if any
	unsigned int vendor_id; // 
	
	char cpi_layer[1024]; // name of the plugin interface used for this device
	char device[1024]; // filename of controlling device, if any // eg. /dev/video/0
		
	unsigned int flags;
};

typedef struct _unicap_device unicap_device_t;

struct _unicap_rect
{
	int x;
	int y;	
	int width;
	int height;
};

typedef struct _unicap_rect unicap_rect_t;

struct _unicap_format
{
      char identifier[128];
      
      // default
      unicap_rect_t size;
      
      // min and max extends
      unicap_rect_t min_size;
      unicap_rect_t max_size;
      
      // stepping: 
      // 0 if no free scaling available
      int h_stepping;
      int v_stepping;
      // array of possible sizes
      unicap_rect_t *sizes;
      int size_count;
      
      int bpp;
      unsigned int fourcc;
      unsigned int flags;

      unsigned int buffer_types; // available buffer types
      int system_buffer_count;

      size_t buffer_size;

      unicap_buffer_type_t buffer_type;
};

typedef struct _unicap_format unicap_format_t;

struct _unicap_data_buffer
{
	unicap_format_t format;
	
	int frame_number;
	struct timeval fill_time; //
	struct timeval duration;
	struct timeval capture_start_time;// optional, time of first copy action of image data of this frame
		
	unsigned char *data;
	size_t buffer_size;	

	unicap_buffer_type_t type; // system pointers or user pointers

	unsigned int reserved[8];		
};

typedef struct _unicap_data_buffer unicap_data_buffer_t;

struct _unicap_property_range
{
	double min;
	double max;
};

typedef struct _unicap_property_range unicap_property_range_t;

struct _unicap_property_value_list
{
	double *values;
	int value_count;	
};

typedef struct _unicap_property_value_list unicap_property_value_list_t;

struct _unicap_property_menu 
{
	char **menu_items;
	int menu_item_count;
};

typedef struct _unicap_property_menu unicap_property_menu_t;

enum _unicap_property_type_enum
{
	UNICAP_PROPERTY_TYPE_RANGE = 0,
	UNICAP_PROPERTY_TYPE_VALUE_LIST,
	UNICAP_PROPERTY_TYPE_MENU,
	UNICAP_PROPERTY_TYPE_DATA,
	UNICAP_PROPERTY_TYPE_FLAGS, 

	UNICAP_PROPERTY_TYPE_UNKNOWN
};

typedef enum _unicap_property_type_enum unicap_property_type_enum_t;

struct _unicap_property
{
	char identifier[128]; //mandatory
	char category[128];
	char unit[128]; // 

	// list of properties identifier which value / behaviour may change if this property changes
	char **relations;
	int relations_count;	
		
	union
	{
		double value; // default if enumerated
		char menu_item[128];
	};
		
	union{	
			unicap_property_range_t range; // if UNICAP_USE_RANGE is asserted
			unicap_property_value_list_t value_list; // if UNICAP_USE_VALUE_LIST is asserted
			unicap_property_menu_t menu;
	};

	double stepping;

	unicap_property_type_enum_t type;	
	u_int64_t flags;        // defaults if enumerated
	u_int64_t flags_mask;	// defines capabilities if enumerated

	// optional data
	void *property_data; 
	size_t property_data_size;	
};

typedef struct _unicap_property unicap_property_t;

#define UNICAP_FLAGS_MANUAL    (1ULL)
#define UNICAP_FLAGS_AUTO      1ULL<<1ULL
#define UNICAP_FLAGS_ONE_PUSH  1ULL<<2ULL
// property may be read during auto
#define UNICAP_FLAGS_READ_OUT  1ULL<<3ULL 
#define UNICAP_FLAGS_ON_OFF    1ULL<<4ULL
#define UNICAP_FLAGS_READ_ONLY 1ULL<<5ULL
#define UNICAP_FLAGS_FORMAT_CHANGE 1ULL<<6ULL

// "system" flags
#define UNICAP_FLAGS_CHECK_STEPPING 1ULL<<32ULL


enum _unicap_event_t
{
	UNICAP_EVENT_FIRST = 0,
	UNICAP_EVENT_DEVICE_REMOVED = 0, 
	UNICAP_EVENT_NEW_DEVICE,
	UNICAP_EVENT_NEW_FRAME, 
	UNICAP_EVENT_DROP_FRAME, 
	UNICAP_EVENT_LAST
};

typedef enum _unicap_event_t unicap_event_t;

typedef struct _unicap_handle *unicap_handle_t;

typedef void (*unicap_callback_t)( unicap_event_t event, ... );
typedef void (*unicap_new_frame_callback_t)( unicap_event_t event, unicap_handle_t handle, unicap_data_buffer_t *buffer, void *user_ptr );
typedef void (*unicap_drop_frame_callback_t)( unicap_event_t event, unicap_handle_t handle, void *user_ptr );
typedef void (*unicap_new_device_callback_t)( unicap_event_t event, unicap_device_t *device, void *user_ptr );

#ifdef __cplusplus
extern "C" {
#endif


// Functions without a handle

/*
  Temporary hack to avoid looping with the remote cpi
*/
unicap_status_t unicap_set_filter_remote( int do_filter );
	

/*
  Rebuild internal list of devices.
  
  Output: 
    count: number of devices found
 */
unicap_status_t unicap_reenumerate_devices( int *count );

/*
 */
unicap_status_t unicap_enumerate_devices( unicap_device_t *specifier, unicap_device_t *device, int index );

/*
  Acquire a handle to a device. 

  Input: 
    device: a device as returned from unicap_enumerate_devices

  Output: 
    handle
 */
unicap_status_t unicap_open( unicap_handle_t *handle, unicap_device_t *device );

//Functions requiring a handle

unicap_status_t unicap_register_callback( unicap_handle_t handle, 
					  unicap_event_t event, 
					  unicap_callback_t func, 
					  void *user_ptr );

/*
  Invalidate the handle. If ref_count == 0, frees all resources associated with that handle
 */
unicap_status_t unicap_close( unicap_handle_t handle );

/*
  Get the device associated with handle
 */
unicap_status_t unicap_get_device( unicap_handle_t handle, unicap_device_t *device );	

/*
  Copy the handle, increment the ref_count
 */
unicap_handle_t unicap_clone_handle( unicap_handle_t old_handle );

/*
  Re-create the list of formats supported by the device. Invalidates all data returned by prior calls 
  to unicap_emumerate_formats() and unicap_get_format(). 

  Output: 
    count: Contains the number of formats currently supported by the device. Might be NULL
 */
unicap_status_t unicap_reenumerate_formats( unicap_handle_t handle, 
					    int *count );

/*
  Enumerate formats known to the device 

  Input: 
    specifier: limits the enumerated formats to the ones matching the fields in specifier. Fields set to -1 in the specifier are ignored
    index:     index of the format in the enumeration

  Output: 
    format: format at index

  Return value: STATUS_SUCCESS
                STATUS_NO_MORE_FORMATS: end of the list of matching formats has been reached
 */
unicap_status_t unicap_enumerate_formats( unicap_handle_t handle, unicap_format_t *specifier, unicap_format_t *format, int index );

/*
  Set a format. 

  Input: 
   format

  Return value: STATUS_SUCCESS
                STATUS_NO_MATCH: given format not present for device
 */
unicap_status_t unicap_set_format( unicap_handle_t handle, unicap_format_t *format );

/*
  Get the current format

  Output: 
    format

  Return value: STATUS_SUCCESS
                STATUS_NO_FORMAT
*/
unicap_status_t unicap_get_format( unicap_handle_t handle, unicap_format_t *format );

/*
  Re-create the list of properties supported by the device. This invalidates all data returned by prior calls to 
  unicap_enumerate_properties() and unicap_get_property()

  Output: 
    count: number of properties supported by the device. Might be NULL
 */
unicap_status_t unicap_reenumerate_properties( unicap_handle_t handle, int *count );

/*
  Enumerate properties matching "specifier"
  
  Input: specifier
         index

  Output: property

  Return value: STATUS_SUCCESS
                STATUS_NO_MATCH
 */
unicap_status_t unicap_enumerate_properties( unicap_handle_t handle, unicap_property_t *specifier, unicap_property_t *property, int index );

/*
  Set a device property
  
  Input: property
  
  Output: STATUS_SUCCESS
          STATUS_NO_MATCH
 */
unicap_status_t unicap_set_property( unicap_handle_t handle, unicap_property_t *property );

/*
  Get a device property
  
  Input: property

  Output: STATUS_SUCCESS
          STATUS_NO_MATCH
 */
unicap_status_t unicap_get_property( unicap_handle_t handle, unicap_property_t *property );


/*
  Start the capture device. After this call, unicap_wait_buffer calls are allowed

  Return value: STATUS_SUCCESS
                STATUS_FAILURE
 */
unicap_status_t unicap_start_capture( unicap_handle_t handle );
/*
  Stop the capture device
 */
unicap_status_t unicap_stop_capture( unicap_handle_t handle );

/*
  Queue a buffer to be filled by the capture device. The queued buffer must not be touched ( especially not be freed ) until
  it is in the ready queue or dequeued. Supplied buffer must be at least of the buffer size returned by get_format

  Input: data_buffer
 */
unicap_status_t unicap_queue_buffer( unicap_handle_t handle, unicap_data_buffer_t *data_buffer );

/*
  Removes the first buffer from the queue. Depending on the device:Can only be called if the capture device is stopped.

  Output: data_buffer: Contains the dequeued buffer or NULL if no buffer was queued
 */
unicap_status_t unicap_dequeue_buffer( unicap_handle_t handle, unicap_data_buffer_t **data_buffer );

/*
  Removes a buffer from the ready queue. If no buffer is available, this function blocks until a buffer got filled.

  Output: data_buffer
 */
unicap_status_t unicap_wait_buffer( unicap_handle_t handle, unicap_data_buffer_t **data_buffer );

/*
  Poll for buffers in the fill queue
 */
unicap_status_t unicap_poll_buffer( unicap_handle_t handle, int *count );


unicap_status_t unicap_describe_device( unicap_device_t *device, char *buffer, size_t *buffer_size );
unicap_status_t unicap_describe_format( unicap_format_t *format, char *buffer, size_t *buffer_size );
unicap_status_t unicap_describe_property( unicap_property_t *property, char *buffer, size_t *buffer_size );

unicap_status_t unicap_void_device( unicap_device_t *device );
unicap_status_t unicap_void_format( unicap_format_t *format );
unicap_status_t unicap_void_property( unicap_property_t *property );


/*
  Makes calling process the stream owner for this device
 */
unicap_status_t unicap_lock_stream( unicap_handle_t handle );
unicap_status_t unicap_unlock_stream( unicap_handle_t handle );
int unicap_is_stream_locked( unicap_device_t *device );
	
/*
  Makes calling process the properties owner
 */
unicap_status_t unicap_lock_properties( unicap_handle_t handle );
	
unicap_status_t unicap_unlock_properties( unicap_handle_t handle );
	
#ifdef __cplusplus
}
#endif


#endif//__UNICAP_H__
