/* unicap
 *
 * Copyright (C) 2004 Arne Caspari ( arne_caspari@users.sourceforge.net )
 *
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <string.h>

#include <sys/time.h>
#include <time.h>

#include "config.h"

#if UNICAPGTK_DEBUG
#define DEBUG
#endif
#include <debug.h>

#include "unicapgtk.h"
#include "unicapgtk_priv.h"
#include "unicapgtk_video_display.h"
#include "display_backends/backend.h"
#include "cs_xfm.h"

#define TIMEOUT_UNSET 0

#define DATA_BUFFERS 8

enum {
   UNICAPGTK_VIDEO_DISPLAY_PREDISPLAY_SIGNAL,
   UNICAPGTK_VIDEO_DISPLAY_POSTDISPLAY_SIGNAL,
   LAST_SIGNAL
};

enum
{
   PROP_0 = 0,
   PROP_SCALE_TO_FIT
};

static gchar * display_backends[] =
{
   "backend_xv", 
   "backend_gtk" 
};


struct _UnicapgtkVideoDisplay
{
      GtkAspectFrame  parent_instance;
      
      GtkWidget *drawing_area;
      GdkGC *gc;
      
      GdkPixbuf *still_image_pixbuf;
      
      gint display_timeout_tag;
      gboolean scale_to_fit;
      gboolean capture_running;
      gboolean pause;
      
      unicap_handle_t unicap_handle;
      unicap_device_t device;
      unicap_format_t format;
      unicap_data_buffer_t data_buffer[DATA_BUFFERS];
      unicap_data_buffer_t still_image_buffer;

      gboolean new_frame;
      
      int crop_x;
      int crop_y;
      int crop_width;
      int crop_height;
      
      int display_width;
      int display_height;

      GModule *backend;
      backend_update_image_func_t backend_update_image;
      backend_redraw_func_t backend_redraw;
      backend_destroy_func_t backend_destroy;
      backend_set_output_size_func_t backend_set_output_size;
      backend_expose_event_t backend_expose;
      backend_get_image_data_t backend_get_image_data;
      backend_lock_t backend_lock;
      backend_unlock_t backend_unlock;
      backend_display_image_t backend_display_image;
      
      gpointer backend_data;
};


static void unicapgtk_video_display_class_init          ( UnicapgtkVideoDisplayClass *klass);
static void unicapgtk_video_display_init                ( UnicapgtkVideoDisplay      *ugtk);
static void unicapgtk_video_display_set_property        ( GObject       *object, 
                                                          guint          property_id, 
                                                          const GValue  *value, 
                                                          GParamSpec    *pspec );
static void unicapgtk_video_display_get_property        ( GObject       *object, 
                                                          guint          property_id, 
                                                          GValue        *value, 
                                                          GParamSpec    *pspec );
static void unicapgtk_video_display_destroy             ( GtkObject     *object );
static void unicapgtk_video_display_compute_child_allocation( GtkFrame  *frame,
                                                              GtkAllocation *child_allocation );
static gboolean tlwindow_configure_event( GtkWidget *window, GdkEventConfigure *event, UnicapgtkVideoDisplay *ugtk );
static void da_hierarchy_changed_cb( GtkWidget *da, GtkWidget *widget2, UnicapgtkVideoDisplay *ugtk );
static void da_configure_cb ( GtkWidget *da, GdkEventConfigure *event, UnicapgtkVideoDisplay *ugtk );
static void da_expose_event( GtkWidget *da, GdkEventExpose *event, UnicapgtkVideoDisplay *ugtk );
static gboolean da_realize_cb( GtkWidget *da, UnicapgtkVideoDisplay *ugtk );
static void set_backend( UnicapgtkVideoDisplay *ugtk );
static void new_frame_cb( unicap_event_t event, unicap_handle_t handle, unicap_data_buffer_t *buffer, UnicapgtkVideoDisplay *ugtk );

static GtkWidgetClass *parent_class = NULL;
static guint unicapgtk_video_display_signals[LAST_SIGNAL] = { 0 };

GType
unicapgtk_video_display_get_type (void)
{
   static GType ugtk_type = 0;

   if( !ugtk_type )
   {
      static const GTypeInfo ugtk_info =
	 {
	    sizeof( UnicapgtkVideoDisplayClass ),
	    NULL, /* base_init */
	    NULL, /* base_finalize */
	    ( GClassInitFunc ) unicapgtk_video_display_class_init,
	    NULL, /* class_finalize */
	    NULL, /* class_data */
	    sizeof( UnicapgtkVideoDisplay ),
	    0,
	    ( GInstanceInitFunc ) unicapgtk_video_display_init,
	 };
	  
      ugtk_type = g_type_register_static( GTK_TYPE_ASPECT_FRAME, "UnicapgtkVideoDisplay", &ugtk_info, 0 );
   }

   return ugtk_type;
}

static void unicapgtk_video_display_class_init( UnicapgtkVideoDisplayClass *klass )
{
   GObjectClass *object_class = G_OBJECT_CLASS( klass );
   GtkObjectClass *gtk_object_class = GTK_OBJECT_CLASS( klass );
   GtkFrameClass *frame_class = GTK_FRAME_CLASS( klass );

   parent_class = g_type_class_peek_parent( klass );

   object_class->set_property = unicapgtk_video_display_set_property;
   object_class->get_property = unicapgtk_video_display_get_property;

   gtk_object_class->destroy  = unicapgtk_video_display_destroy;

   frame_class->compute_child_allocation = unicapgtk_video_display_compute_child_allocation;

   unicapgtk_video_display_signals[UNICAPGTK_VIDEO_DISPLAY_PREDISPLAY_SIGNAL] = 
      g_signal_new( "unicapgtk_video_display_predisplay",
		    G_TYPE_FROM_CLASS(klass),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
		    G_STRUCT_OFFSET(UnicapgtkVideoDisplayClass, unicapgtk_video_display),
		    NULL, 
		    NULL,                
		    g_cclosure_marshal_VOID__POINTER,
		    G_TYPE_NONE, 1, 
		    G_TYPE_POINTER );

   unicapgtk_video_display_signals[UNICAPGTK_VIDEO_DISPLAY_POSTDISPLAY_SIGNAL] = 
      g_signal_new( "unicapgtk_video_display_postdisplay",
		    G_TYPE_FROM_CLASS(klass),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
		    G_STRUCT_OFFSET(UnicapgtkVideoDisplayClass, unicapgtk_video_display),
		    NULL, 
		    NULL,                
		    g_cclosure_marshal_VOID__POINTER,
		    G_TYPE_NONE, 1,
		    G_TYPE_POINTER );

/*    unicapgtk_video_display_signals[UNICAPGTK_VIDEO_DISPLAY_DISPLAY_PIXBUF_SIGNAL] =  */
/*       g_signal_new( "unicapgtk_video_display_display_pixbuf", */
/* 		    G_TYPE_FROM_CLASS(klass), */
/* 		    G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION, */
/* 		    G_STRUCT_OFFSET(UnicapgtkVideoDisplayClass, unicapgtk_video_display), */
/* 		    NULL,  */
/* 		    NULL,                 */
/* 		    g_cclosure_marshal_VOID__POINTER, */
/* 		    G_TYPE_NONE, 1, */
/* 		    G_TYPE_POINTER ); */


   g_object_class_install_property (object_class,
				    PROP_SCALE_TO_FIT,
				    g_param_spec_boolean ("scale-to-fit",
                                                          NULL, NULL,
                                                          FALSE,
                                                          G_PARAM_READWRITE));
}

static void unicapgtk_video_display_set_property( GObject *object, 
						  guint property_id, 
						  const GValue *value, 
						  GParamSpec *pspec )
{
   UnicapgtkVideoDisplay *ugtk = UNICAPGTK_VIDEO_DISPLAY( object );

   switch( property_id )
   {
      case PROP_SCALE_TO_FIT:
	 unicapgtk_video_display_set_scale_to_fit( ugtk,
						   g_value_get_boolean( value ) );
	 break;

      default:
	 G_OBJECT_WARN_INVALID_PROPERTY_ID( object, property_id, pspec );
	 break;
   }
}

static void unicapgtk_video_display_get_property( GObject *object, 
						  guint property_id, 
						  GValue *value, 
						  GParamSpec *pspec )
{
   UnicapgtkVideoDisplay *ugtk = UNICAPGTK_VIDEO_DISPLAY( object );

   switch( property_id )
   {
      case PROP_SCALE_TO_FIT:
         g_value_set_boolean( value, ugtk->scale_to_fit );
         break;

      default:
	 G_OBJECT_WARN_INVALID_PROPERTY_ID( object, property_id, pspec );
	 break;
   }
}


static void unicapgtk_video_display_destroy( GtkObject *object )
{
   UnicapgtkVideoDisplay *ugtk = UNICAPGTK_VIDEO_DISPLAY( object );

   if( ugtk->capture_running )
   {
      unicapgtk_video_display_stop( ugtk );
   }

   if( ugtk->unicap_handle )
   {
      unicap_close( ugtk->unicap_handle );
      ugtk->unicap_handle = NULL;
   }

   if( ugtk->backend )
   {
      ugtk->backend_destroy( ugtk->backend_data );
      g_module_close( ugtk->backend );
      ugtk->backend = NULL;
   }

   GTK_OBJECT_CLASS (parent_class)->destroy (object);
}


static void unicapgtk_video_display_compute_child_allocation( GtkFrame  *frame, GtkAllocation *child_allocation )
{
   UnicapgtkVideoDisplay *ugtk = UNICAPGTK_VIDEO_DISPLAY( frame );
   
   GTK_FRAME_CLASS( parent_class )->compute_child_allocation( frame, child_allocation );
   
   if( ugtk->scale_to_fit )
   {
      return;
   }

   if( child_allocation->width > ugtk->crop_width )
   {
      child_allocation->x += GTK_ASPECT_FRAME( frame )->xalign *
	 (child_allocation->width - ugtk->crop_width);
      child_allocation->width = ugtk->crop_width;
   }
   if( child_allocation->height > ugtk->crop_height )
   {
      child_allocation->y += GTK_ASPECT_FRAME( frame )->yalign *
	 (child_allocation->height - ugtk->crop_height);
      child_allocation->height = ugtk->crop_height;
   }
}

static 
void da_size_allocate( GtkWidget *da, GtkAllocation *allocation, UnicapgtkVideoDisplay * ugtk )
{
   if( ugtk->scale_to_fit )
   {
      if( ugtk->backend_set_output_size )
      {
	 ugtk->backend_set_output_size( ugtk->backend_data, allocation->width, allocation->height );
      }
      else
      {
	 {
	    g_warning( "Backend does not support scaling\n" );
	 }
      }
   }
}


static 
void unicapgtk_video_display_init( UnicapgtkVideoDisplay *ugtk )
{
   gtk_aspect_frame_set( GTK_ASPECT_FRAME( ugtk ), 0.5, 0.5, 1.0, TRUE );
   gtk_frame_set_shadow_type( GTK_FRAME( ugtk ), GTK_SHADOW_NONE );

   ugtk->drawing_area = gtk_drawing_area_new( );
   gtk_widget_add_events( ugtk->drawing_area, 
			  GDK_ALL_EVENTS_MASK );
   gtk_container_add( GTK_CONTAINER( ugtk ), ugtk->drawing_area );	
   gtk_widget_show( ugtk->drawing_area );


   g_signal_connect( ugtk->drawing_area, "hierarchy-changed",
		     G_CALLBACK( da_hierarchy_changed_cb ), ugtk );
   g_signal_connect( ugtk->drawing_area, "configure-event",
		     G_CALLBACK( da_configure_cb ),
                     ugtk );
   g_signal_connect( ugtk->drawing_area, "unmap",
                     G_CALLBACK(gtk_widget_destroyed),
                     &ugtk->gc );

   g_signal_connect( ugtk->drawing_area, "expose-event", 
		     G_CALLBACK( da_expose_event ), 
		     ugtk );
   g_signal_connect( ugtk->drawing_area, "size-allocate", 
		     G_CALLBACK( da_size_allocate ), 
		     ugtk );

   ugtk->unicap_handle = NULL;  
   ugtk->scale_to_fit = FALSE;
   ugtk->display_timeout_tag = TIMEOUT_UNSET;

}

/*
  Initializes object structures. Looks for first matching video format and resizes the drawing area accordingly.

  Input: ugtk
  
  Output: 

  Returns:
*/
static void new( UnicapgtkVideoDisplay *ugtk )
{
   unicap_format_t format_spec, format;
   unicap_void_format( &format_spec );
   format_spec.fourcc = UYVY;
	
   ugtk->capture_running = FALSE;
   ugtk->pause = FALSE;
   ugtk->display_timeout_tag = TIMEOUT_UNSET;
   ugtk->still_image_pixbuf = NULL;
   ugtk->backend = NULL;   

   if( !SUCCESS( unicap_enumerate_formats( ugtk->unicap_handle, 
					   &format_spec, 
					   &format, 
					   0 ) ) )
   {
      if( !SUCCESS( unicap_enumerate_formats( ugtk->unicap_handle, 
					      NULL, 
					      &format, 
					      0 ) ) )
      {
	 TRACE( "Failed to get video format\n" );
	 return;
      }
   }
	
   if( ( format.size.width == 0 ) || 
       ( format.size.height == 0 ) )
   {
      if( format.size_count )
      {
	 format.size.width = format.sizes[ format.size_count - 1 ].width;
	 format.size.height = format.sizes[ format.size_count -1 ].height;
      }
   }


   if( !unicapgtk_video_display_set_format( ugtk, &format ) )
   {
      TRACE( "Failed to set video format\n" );
      return;
   }
   if( !SUCCESS( unicap_get_format( ugtk->unicap_handle, 
				    &ugtk->format ) ) )
   {
      TRACE( "Failed to get video format\n" );
      return;
   }

   TRACE( "format: FOURCC = %c%c%c%c\n", format.fourcc, format.fourcc>>8, format.fourcc>>16, 
	  format.fourcc>>24 );

}

/*
 */
GtkWidget* unicapgtk_video_display_new( void )
{
   return g_object_new( UNICAPGTK_VIDEO_DISPLAY_TYPE, NULL );
}

/*
  Creates new video display widget by device spec. 

  Input: device_spec : Specification or NULL for any device
*/
GtkWidget* unicapgtk_video_display_new_by_device( unicap_device_t *device_spec )
{
   UnicapgtkVideoDisplay *ugtk;

   ugtk = g_object_new( UNICAPGTK_VIDEO_DISPLAY_TYPE, NULL );

   if( SUCCESS( unicap_enumerate_devices( device_spec, &ugtk->device, 0 ) ) )
   {
      unicap_open( &ugtk->unicap_handle, &ugtk->device );
   }
	
   if( ugtk->unicap_handle )
   {
      new( ugtk );
   }

   return GTK_WIDGET( ugtk );
}

/*
  Creates new video display widget by existing handle.

  Input: handle : initialized unicap handle
*/
GtkWidget *unicapgtk_video_display_new_by_handle( unicap_handle_t handle )
{
   UnicapgtkVideoDisplay *ugtk;

   ugtk = g_object_new( UNICAPGTK_VIDEO_DISPLAY_TYPE, NULL );

   ugtk->unicap_handle = unicap_clone_handle( handle );
	
   new( ugtk );
	
   return GTK_WIDGET( ugtk );
}


static void da_hierarchy_changed_cb( GtkWidget *da, GtkWidget *widget2, UnicapgtkVideoDisplay *ugtk )
{
   GtkWidget *window;
   
   window = gtk_widget_get_toplevel( da );
   if( window )
   {
      g_signal_connect( G_OBJECT( window ), "configure-event", G_CALLBACK( tlwindow_configure_event ), ugtk );
   }
}

static gboolean tlwindow_configure_event( GtkWidget *window, GdkEventConfigure *event, UnicapgtkVideoDisplay *ugtk )
{
   gtk_widget_queue_draw( GTK_WIDGET( ugtk ) );
   return FALSE;
}

static void da_configure_cb ( GtkWidget *da, GdkEventConfigure *event, UnicapgtkVideoDisplay *ugtk )
{
   if( ugtk->scale_to_fit )
   {
      unicapgtk_video_display_set_size( ugtk, event->width, event->height );
   }
}

static void da_expose_event( GtkWidget *da, GdkEventExpose *event, UnicapgtkVideoDisplay *ugtk )
{
   unicap_data_buffer_t data_buffer;

   if( !ugtk->backend )
   {
      return;
   }

   ugtk->backend_lock( ugtk->backend_data );
   
   if( ugtk->new_frame )
   {
      ugtk->backend_get_image_data( ugtk->backend_data, &data_buffer, 1 );
      
      g_signal_emit( G_OBJECT( ugtk ), 
		     unicapgtk_video_display_signals[ UNICAPGTK_VIDEO_DISPLAY_PREDISPLAY_SIGNAL], 
		     0, 
		     (gpointer)&data_buffer );
      ugtk->backend_display_image( ugtk->backend_data );
   }

   ugtk->backend_expose( ugtk->backend_data, da, event );

   if( ugtk->new_frame )
   {

      g_signal_emit( G_OBJECT( ugtk ), 
		     unicapgtk_video_display_signals[ UNICAPGTK_VIDEO_DISPLAY_POSTDISPLAY_SIGNAL], 
		     0, 
		     (gpointer)&data_buffer );
      ugtk->new_frame = FALSE;
   }

   ugtk->backend_unlock( ugtk->backend_data );


}


static gint timeout_cb( UnicapgtkVideoDisplay *ugtk )
{
   if( ugtk->pause )
   {
      return TRUE;
   }

   if( ugtk->new_frame )
   {
      gdk_threads_enter();
      gtk_widget_queue_draw( GTK_WIDGET( ugtk ) );
      gdk_threads_leave();
   }

   return TRUE;
}

static void new_frame_cb( unicap_event_t event, unicap_handle_t handle, unicap_data_buffer_t *buffer, UnicapgtkVideoDisplay *ugtk )
{
   ugtk->new_frame = TRUE;
   if( ugtk->backend )
   {
      ugtk->backend_lock( ugtk->backend_data );
      ugtk->backend_update_image( ugtk->backend_data, buffer, NULL );
      ugtk->backend_unlock( ugtk->backend_data );
   }
}


/*
  Start capturing on the current device and display the video stream
*/
gboolean unicapgtk_video_display_start( UnicapgtkVideoDisplay *ugtk )
{
   int cbuf;
   unicap_format_t tmp_format;
   
   if( !ugtk->unicap_handle )
   {
      g_warning( "Implement me: Correct warning message - handle not set\n" );
      return FALSE;
   }

   if( ugtk->capture_running )
   {
      g_warning( "video_display_start: Display already started!\n" );
      return FALSE;
   }

   if( ugtk->display_timeout_tag != TIMEOUT_UNSET )
   {
      g_source_remove( ugtk->display_timeout_tag );
      ugtk->display_timeout_tag = TIMEOUT_UNSET;
   }

   if( ugtk->still_image_buffer.data )   //
   {
      free( ugtk->still_image_buffer.data );
      ugtk->still_image_buffer.data = NULL;
   }

   if( !strcmp( ugtk->format.identifier, "" ) )
   {
      g_warning( "Video format not set" );
      return FALSE;
   }


   memset( &ugtk->still_image_buffer, 0x0, sizeof( unicap_data_buffer_t ) );
   ugtk->still_image_buffer.data = malloc( ugtk->format.buffer_size );
   ugtk->still_image_buffer.buffer_size = ugtk->format.buffer_size;
   g_memmove( &ugtk->still_image_buffer.format, &ugtk->format, sizeof( unicap_format_t ) );

   ugtk->capture_running = TRUE;

   if( !SUCCESS( unicap_get_format( ugtk->unicap_handle, &tmp_format ) ) )
   {
      g_warning( "Failed to get current video format\n" );
      return FALSE;
   }
   
   tmp_format.buffer_type = UNICAP_BUFFER_TYPE_SYSTEM;
   
   if( !SUCCESS( unicap_set_format( ugtk->unicap_handle, &tmp_format ) ) )
   {
      g_warning( "Failed to activate SYSTEM_BUFFERS\n" );
      return FALSE;
   }   

   unicap_register_callback( ugtk->unicap_handle, UNICAP_EVENT_NEW_FRAME, new_frame_cb, ugtk );

   ugtk->display_timeout_tag = g_timeout_add( 30, (GtkFunction)timeout_cb, ugtk );

   if( !SUCCESS( unicap_start_capture( ugtk->unicap_handle ) ) )
   {
      g_warning( "Failed to start video capture\n" );
      return FALSE;
   }
   

   return TRUE;
}

/*
  Stop capturing on current device
*/
void unicapgtk_video_display_stop( UnicapgtkVideoDisplay *ugtk )
{
   int cbuf;
   if( !ugtk->capture_running )
   {
/*       g_warning( "video_display not running\n" ); */
      return;
   }

   ugtk->capture_running = FALSE;

   if( ugtk->display_timeout_tag != TIMEOUT_UNSET )
   {
      g_source_remove( ugtk->display_timeout_tag );
      ugtk->display_timeout_tag = TIMEOUT_UNSET;
   }
	
   unicap_stop_capture( ugtk->unicap_handle );
   

   if( ugtk->still_image_buffer.data )   //
   {
      free( ugtk->still_image_buffer.data );
      ugtk->still_image_buffer.data = NULL;
   }
	
   for( cbuf = 0; cbuf < DATA_BUFFERS; cbuf++ )
   {
      if( ugtk->data_buffer[cbuf].type == UNICAP_BUFFER_TYPE_USER )
      {
	 g_free( ugtk->data_buffer[cbuf].data );
      }
   }
   
   // todo: redraw current image as RGB buffer
}

/*
  Get current device
*/
unicap_device_t *unicapgtk_video_display_get_device( UnicapgtkVideoDisplay *ugtk )
{
   return &ugtk->device;
}

/*
  Get current handle
*/
unicap_handle_t unicapgtk_video_display_get_handle( UnicapgtkVideoDisplay *ugtk )
{
   return unicap_clone_handle( ugtk->unicap_handle );
}


/*
  Change video format
*/
gboolean unicapgtk_video_display_set_format( UnicapgtkVideoDisplay *ugtk, 
					     unicap_format_t *format_spec )
{
   unicap_format_t format;
   gboolean old_capture_state;
   gboolean status = TRUE;
   

   old_capture_state = ugtk->capture_running;
   if( ugtk->capture_running )
   {
      unicapgtk_video_display_stop( ugtk );
   }

   if( ugtk->backend )
   {
      ugtk->backend_destroy( ugtk->backend_data );
      g_module_close( ugtk->backend );
      ugtk->backend = NULL;
   }

   if( ugtk->unicap_handle )
   {
      if( !SUCCESS( unicap_enumerate_formats( ugtk->unicap_handle, 
					      format_spec, 
					      &format, 
					      0 ) ) )
      {
	 TRACE( "Failed to enumerate video format\n" );
	 return FALSE;
      }
	
      format.size.width = format_spec->size.width;
      format.size.height = format_spec->size.height;

      if( !SUCCESS( unicap_set_format( ugtk->unicap_handle, &format ) ) )
      {
	 TRACE( "Failed to set video format\n" );
      }
      unicap_get_format( ugtk->unicap_handle, &ugtk->format );
   }
   else
   {
      unicap_copy_format( &format, format_spec );
      unicap_copy_format( &ugtk->format, format_spec );
   }

   TRACE( "format: FOURCC = %c%c%c%c size: %d x %d\n", format.fourcc, format.fourcc>>8, format.fourcc>>16, 
	  format.fourcc>>24, 
	  format.size.width, 
	  format.size.height);


   ugtk->crop_x = ugtk->crop_y = 0;
   ugtk->crop_width = format.size.width;
   ugtk->crop_height = format.size.height;

   ugtk->display_width = format.size.width;
   ugtk->display_height = format.size.height;

   if ( ugtk->scale_to_fit)
   {
      if( format.size.width > 0 && format.size.height > 0 )
      {
	 gtk_aspect_frame_set( GTK_ASPECT_FRAME( ugtk ), 0.5, 0.5,
			       (gdouble) format.size.width /
			       (gdouble) format.size.height,
			       FALSE );
      }

      gtk_widget_queue_resize( GTK_WIDGET( ugtk->drawing_area ) );
   }
   else
   {
      gtk_widget_set_size_request( GTK_WIDGET( ugtk->drawing_area ),
                                   format.size.width,
                                   format.size.height );
   }

   if( GTK_WIDGET_REALIZED( ugtk->drawing_area ) )
   {
      set_backend( ugtk );
   }
   else
   {
      g_signal_connect_after( G_OBJECT( ugtk->drawing_area ), "realize", G_CALLBACK(da_realize_cb), ugtk );
   }

   if( old_capture_state )
   {
      status = unicapgtk_video_display_start( ugtk );
   }

   return status;
}

void unicapgtk_video_display_get_format( UnicapgtkVideoDisplay *ugtk, unicap_format_t *format )
{
   unicap_void_format( format );
   if( ugtk->unicap_handle )
   {
      unicap_get_format( ugtk->unicap_handle, format );
   }
}

/*
  stop capturing images ( but do not stop the device );
  set current image as still image
*/
void unicapgtk_video_display_set_pause( UnicapgtkVideoDisplay *ugtk, gboolean state )
{
   ugtk->pause = state ? TRUE : FALSE;

   gtk_widget_queue_draw( ugtk->drawing_area );
}


/*
  get pause state
*/
gboolean unicapgtk_video_display_get_pause( UnicapgtkVideoDisplay *ugtk )
{
   return ugtk->pause;
}

/*
  set crop window
  
  crop_x : start x position in source buffer
  crop_y : start y position in source buffer
  crop_width : width of crop window
  crop_height : height of crop window
*/
void unicapgtk_video_display_set_crop( UnicapgtkVideoDisplay *ugtk, 
				       UnicapgtkVideoDisplayCropping *crop )
{
   crop->crop_x = MAX( crop->crop_x, 0 );
   crop->crop_x = MIN( crop->crop_x, ugtk->format.size.width );
   crop->crop_width = MAX( crop->crop_width, 0 );
   crop->crop_width = MIN( crop->crop_width, ugtk->format.size.width );

   crop->crop_y = MAX( crop->crop_y, 0 );
   crop->crop_y = MIN( crop->crop_y, ugtk->format.size.height );
   crop->crop_height = MAX( crop->crop_height, 0 );
   crop->crop_height = MIN( crop->crop_height, ugtk->format.size.height );

   ugtk->crop_x = crop->crop_x;
   ugtk->crop_y = crop->crop_y;
   ugtk->crop_width = crop->crop_width;
   ugtk->crop_height = crop->crop_height;

   if (! ugtk->scale_to_fit)
   {
      gtk_widget_set_size_request( ugtk->drawing_area, 
                                   crop->crop_width, 
                                   crop->crop_height );
   }
}

void unicapgtk_video_display_get_crop( UnicapgtkVideoDisplay *ugtk, 
				       UnicapgtkVideoDisplayCropping *crop )
{
   crop->crop_x = ugtk->crop_x;
   crop->crop_y = ugtk->crop_y;
   crop->crop_width = ugtk->crop_width;
   crop->crop_height = ugtk->crop_height;
}


/*
  change device
*/
gboolean unicapgtk_video_display_set_device( UnicapgtkVideoDisplay *ugtk, unicap_device_t *device_spec )
{	
   unicap_status_t status;
   gboolean capture_running;

   capture_running = ugtk->capture_running;

   unicapgtk_video_display_stop( ugtk );

   if( ugtk->unicap_handle )
   {
      unicap_close( ugtk->unicap_handle );
      ugtk->unicap_handle = 0;
   }

   status = unicap_enumerate_devices( device_spec, &ugtk->device, 0 );
   if( SUCCESS( status ) )
   {
      status = unicap_open( &ugtk->unicap_handle, &ugtk->device );
   }	
	
   if( ugtk->unicap_handle )
   {
      new( ugtk );
   }
   else
   {
      TRACE( "FAILED!\n" );
   }

   if( capture_running )
   {
      unicapgtk_video_display_start( ugtk );
   }
	
   return SUCCESS( status );
}

gboolean unicapgtk_video_display_set_handle( UnicapgtkVideoDisplay *ugtk, unicap_handle_t handle )
{	
   unicap_status_t status;
   gboolean capture_running;

   capture_running = ugtk->capture_running;
   if( ugtk->capture_running )
   {
      unicapgtk_video_display_stop( ugtk );
   }

   if( ugtk->unicap_handle )
   {
      unicap_close( ugtk->unicap_handle );
      ugtk->unicap_handle = 0;
   }
	
   ugtk->unicap_handle = unicap_clone_handle( handle );
   status = unicap_get_device( handle, &ugtk->device );
	
   if( ugtk->unicap_handle )
   {
      new( ugtk );
   }
   else
   {
      TRACE( "FAILED!\n" );
   }

   if( capture_running )
   {
      unicapgtk_video_display_start( ugtk );
   }
	
   return SUCCESS( status );
}

GdkPixbuf *unicapgtk_video_display_get_still_image( UnicapgtkVideoDisplay *ugtk )
{
   unicap_data_buffer_t *returned_buffer;
   if( SUCCESS( unicap_wait_buffer( ugtk->unicap_handle, &returned_buffer ) ) )
   {
      ugtk->still_image_buffer.type = UNICAP_BUFFER_TYPE_USER;
      memcpy( ugtk->still_image_buffer.data, returned_buffer->data, ugtk->still_image_buffer.format.buffer_size );
      if( !SUCCESS( unicap_queue_buffer( ugtk->unicap_handle, returned_buffer ) ) )
      {
	 TRACE( "Failed to queue buffer!\n" );
      }
      TRACE( "Implement me: drop frames\n" );
   }

   return cs_xfm_new_allocated_pixbuf( &ugtk->still_image_buffer );
}

void unicapgtk_video_display_set_still_image( UnicapgtkVideoDisplay *ugtk, unicap_data_buffer_t *buffer )
{
   if( !ugtk->pause )
   {
      unicapgtk_video_display_set_pause( ugtk, TRUE );
   }

   if( buffer )
   {
      
      ugtk->backend_update_image( ugtk->backend_data, buffer, NULL );
   }

   gtk_widget_queue_draw( GTK_WIDGET(ugtk) );
}

void unicapgtk_video_display_set_size( UnicapgtkVideoDisplay *ugtk, 
				       gint width, gint height )
{
   ugtk->display_width = width;
   ugtk->display_height = height;
}

void unicapgtk_video_display_set_scale_to_fit( UnicapgtkVideoDisplay *ugtk,
                                               gboolean               scale )
{
   g_return_if_fail( IS_UNICAPGTK_VIDEO_DISPLAY( ugtk ) );

   if( scale == ugtk->scale_to_fit )
      return;

   ugtk->scale_to_fit = scale ? TRUE : FALSE;

   if ( ugtk->scale_to_fit )
   {
      unicapgtk_video_display_set_size( ugtk,
                                        GTK_WIDGET (ugtk->drawing_area)->allocation.width,
                                        GTK_WIDGET (ugtk->drawing_area)->allocation.height );
   }
   else
   {
      gtk_widget_set_size_request( GTK_WIDGET( ugtk->drawing_area ),
                                   ugtk->crop_width, ugtk->crop_height );
      gtk_aspect_frame_set( GTK_ASPECT_FRAME( ugtk ), 0.5, 0.5, 1.0, TRUE );
   }

   g_object_notify( G_OBJECT( ugtk ), "scale-to-fit" );
}


static gboolean da_realize_cb( GtkWidget *da, UnicapgtkVideoDisplay *ugtk )
{
   set_backend( ugtk );

   return FALSE;
   
}


static void set_backend( UnicapgtkVideoDisplay *ugtk )
{   
   gboolean res;
   int i;
   gint merit = 0;
   GKeyFile *keyfile;
   gchar *keyfile_path;
   gchar *userconf_path;

   userconf_path = unicapgtk_get_user_config_path( );
   keyfile_path = g_build_path( G_DIR_SEPARATOR_S, userconf_path, "unicapgtk.conf", NULL );
   g_free( userconf_path );
   if( !g_file_test( keyfile_path, G_FILE_TEST_EXISTS ) )
   {
      g_free( keyfile_path );
      keyfile_path = g_build_path( G_DIR_SEPARATOR_S, SYSCONFDIR, "unicap", "unicapgtk.conf", NULL );
   }
   
   keyfile = g_key_file_new();
   if( keyfile )
   {
      if( !g_key_file_load_from_file( keyfile, keyfile_path, 0, NULL ) )
      {
	 g_key_file_free( keyfile );
	 keyfile = NULL;
      }      
   }

   g_free( keyfile_path );
   
   if( ugtk->backend )
   {
      ugtk->backend_destroy( ugtk->backend_data );
      g_module_close( ugtk->backend );
      ugtk->backend = NULL;
   }      

   ugtk->backend_destroy = NULL;
   ugtk->backend_update_image = NULL;
   ugtk->backend_redraw = NULL;
   ugtk->backend_set_output_size = NULL;
   ugtk->backend_expose = NULL;

   for( i = 0; i < ( sizeof( display_backends ) / sizeof( display_backends[0] ) ); i++ )
   {
      GModule *module;
      backend_init_func_t backend_init;
      backend_get_merit_func_t get_merit;
      gchar *path;
      gchar *plugin_path;
      gint module_merit = -1;

      plugin_path = g_build_path( G_DIR_SEPARATOR_S, PKGLIBDIR, "backends", NULL );
      path = g_module_build_path( plugin_path, display_backends[i] );
      module = g_module_open( path, G_MODULE_BIND_LOCAL );
      g_free( path );
      g_free( plugin_path );

      if( !module )
      {
	 TRACE( "Failed to open backend: %s: %s\n", path, g_module_error() );
	 continue;
      }
      
      res = g_module_symbol( module, "backend_get_merit", (gpointer)&get_merit );
      g_assert( res );
      module_merit = get_merit();
      if( keyfile )
      {
	 gchar *val;
	 gchar *modname;
	 modname = g_path_get_basename( g_module_name( module ) );
	 TRACE( "Check keyfile for: %s\n", modname );
	 val = g_key_file_get_value( keyfile, "video_display_be_merit", modname, NULL );
	 g_free( modname );
	 
	 if( val )
	 {
	    module_merit = atoi( val );
	 }
      }

      
      TRACE( "[%s] : %d\n", g_module_name( module ), module_merit );
      if( module_merit > merit )
      {
	 gpointer backend_data;
	 
	 res = g_module_symbol( module, "backend_init", (gpointer)&backend_init );
	 g_assert( res );
	 if( backend_init( ugtk->drawing_area, &ugtk->format, &backend_data, NULL ) )
	 {
	    if( ugtk->backend )
	    {
	       ugtk->backend_destroy( ugtk->backend_data );
	       g_module_close( ugtk->backend );
	       ugtk->backend = NULL;
	    }
	    
	    res = g_module_symbol( module, "backend_destroy", (gpointer)&ugtk->backend_destroy );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_update_image", (gpointer)&ugtk->backend_update_image );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_redraw", (gpointer)&ugtk->backend_redraw );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_set_output_size", (gpointer)&ugtk->backend_set_output_size );
	    res = g_module_symbol( module, "backend_expose_event", (gpointer)&ugtk->backend_expose );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_get_image_data", (gpointer)&ugtk->backend_get_image_data );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_lock", (gpointer)&ugtk->backend_lock );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_unlock", (gpointer)&ugtk->backend_unlock );
	    g_assert( res );
	    res = g_module_symbol( module, "backend_display_image", (gpointer)&ugtk->backend_display_image );
	    g_assert( res );

	    ugtk->backend = module;
	    ugtk->backend_data = backend_data;
	    
	    merit = module_merit;
	 }
	 else
	 {
	    g_module_close( module );
	 }

      }
      else
      {
	 g_module_close( module );
      }
   }
   if( keyfile )
   {
      g_key_file_free( keyfile );
   }
   
   g_assert( ugtk->backend );

   TRACE( "backend: %s\n", g_module_name( ugtk->backend ) );

}
